import datetime
import os
import sys
import traceback
from os.path import join, isfile

import aspen
from aspen import dispatcher, resources, sockets
from aspen.http.request import Request
from aspen.http.response import Response
from aspen.configuration import Configurable
from aspen.utils import to_rfc822, utc

# 2006-11-17 was the first release of aspen - v0.3
THE_PAST = to_rfc822(datetime.datetime(2006, 11, 17, tzinfo=utc))


class Website(Configurable):
    """Represent a website.

    This object holds configuration information, and also knows how to start
    and stop a server, *and* how to handle HTTP requests (per WSGI). It is
    available to user-developers inside of their resources and hooks.

    """

    def __init__(self, argv=None):
        """Takes an argv list, without the initial executable name.
        """
        self.configure(argv)


    def __call__(self, environ, start_response):
        """WSGI interface.
        """
        request = Request.from_wsgi(environ) # too big to fail :-/
        request.website = self
        response = self.handle_safely(request)
        response.request = request # Stick this on here at the last minute
                                   # in order to support close hooks.
        return response(environ, start_response)


    # Interface for Server
    # ====================

    def start(self):
        aspen.log_dammit("Starting up Aspen website.")
        self.hooks.run('startup', self)
        self.network_engine.start()

    def stop(self):
        aspen.log_dammit("Shutting down Aspen website.")
        self.hooks.run('shutdown', self)
        self.network_engine.stop()


    # Request Handling
    # ================

    def handle_safely(self, request):
        """Given an Aspen request, return an Aspen response.
        """
        try:
            request = self.do_inbound(request)
            response = self.handle(request)
        except:
            response = self.handle_error(request)

        response = self.do_outbound(response)
        return response


    def handle(self, request):
        """Given an Aspen request, return an Aspen response.

        By default we use Resource subclasses to generate responses from
        simplates on the filesystem. See aspen/resources/__init__.py.

        You can monkey-patch this method to implement single-page web apps or
        other things in configure-aspen.py:

            from aspen import Response

            def greetings_program(website, request):
                return Response(200, "Greetings, program!")

            website.handle = greetings_program

        Unusual but allowed.

        """
        # Look for a Socket.IO socket (http://socket.io/).
        if isinstance(request.socket, Response):    # handshake
            response = request.socket
            request.socket = None
        elif request.socket is None:                # non-socket
            request.resource = resources.get(request)
            response = request.resource.respond(request)
        else:                                       # socket
            response = request.socket.respond(request)
        response.request = request
        return response


    # Inbound
    # =======

    def do_inbound(self, request):
        request = self.hooks.run('inbound_early', request)
        request = self.hooks.run('inbound_core', request)
        request = self.hooks.run('inbound_late', request)
        return request

    def reset_inbound_core(self):
        self.hooks.inbound_core = [ self.set_fs_etc
                                  , self.set_socket
                                   ]

    def set_fs_etc(self, request):
        dispatcher.dispatch(request)  # mutates request

    def set_socket(self, request):
        request.socket = sockets.get(request)


    # Error
    # =====

    def handle_error(self, request):
        """Given a request, return a response.
        """
        try:                        # nice error messages
            tb_1 = self.log_error()
            request = self.hooks.run('error_early', request)
            response = self.handle_error_nicely(tb_1, request)
        except Response, response:  # error simplate raised Response
            pass
        except:                     # last chance for tracebacks in the browser
            response = self.handle_error_at_all(tb_1)

        response.request = request
        response = self.hooks.run('error_late', response)
        return response


    def log_error(self):
        tb_1 = traceback.format_exc()
        aspen.log_dammit(tb_1)
        return tb_1


    def handle_error_nicely(self, tb_1, request):

        response = sys.exc_info()[1]

        if not isinstance(response, Response):

            # We have a true Exception; convert it to a Response object.

            response = Response(500, tb_1)
            response.request = request

        if 200 <= response.code < 300:

            # The app raised a Response(2xx). Act as if nothing
            # happened. This is unusual but allowed.

            pass

        else:

            # Delegate to any error simplate.
            # ===============================

            fs = self.ours_or_theirs(str(response.code) + '.html')
            if fs is None:
                fs = self.ours_or_theirs('error.html')
            if fs is not None:
                request.fs = fs
                request.original_resource = request.resource
                request.resource = resources.get(request)
                response = request.resource.respond(request, response)

        return response


    def handle_error_at_all(self, tb_1):
        tb_2 = traceback.format_exc().strip()
        tbs = '\n\n'.join([tb_2, "... while handling ...", tb_1])
        aspen.log_dammit(tbs)
        if self.show_tracebacks:
            response = Response(500, tbs)
        else:
            response = Response(500)
        return response


    # Outbound
    # ========

    def do_outbound(self, response):
        response = self.hooks.run('outbound', response)
        return response

    def reset_outbound(self):
        self.hooks.outbound = [self.log_access]

    def log_access(self, response):
        """Log access. With our own format (not Apache's).
        """

        if self.logging_threshold > 0: # short-circuit
            return


        # What was the URL path translated to?
        # ====================================

        fs = getattr(response.request, 'fs', '')
        if fs.startswith(self.www_root):
            fs = fs[len(self.www_root):]
            if fs:
                fs = '.'+fs
        else:
            fs = '...' + fs[-21:]
        msg = "%-24s %s" % (response.request.line.uri.path.raw, fs)


        # Where was response raised from?
        # ===============================

        filename, linenum = response.whence_raised()
        if filename is not None:
            response = "%s (%s:%d)" % (response, filename, linenum)
        else:
            response = str(response)

        # Log it.
        # =======

        aspen.log("%-36s %s" % (response, msg))


    # File Resolution
    # ===============

    def find_ours(self, filename):
        """Given a filename, return a filepath.
        """
        return join(os.path.dirname(__file__), 'www', filename)

    def ours_or_theirs(self, filename):
        """Given a filename, return a filepath or None.
        """
        if self.project_root is not None:
            theirs = join(self.project_root, filename)
            if isfile(theirs):
                return theirs

        ours = self.find_ours(filename)
        if isfile(ours):
            return ours

        return None


    # Conveniences for testing
    # ========================
    # XXX Sure seems like this class should be refactored so we use the same
    # code for both testing and production here.

    def serve_request(self, path):
        """Given an URL path, return response.
        """
        request = Request(uri=path)
        request.website = self
        response = self.handle_safely(request)
        return response


    def load_simplate(self, path, request=None, return_request_too=False):
        """Given an URL path, return a simplate (Resource) object.
        """
        if request is None:
            request = Request(uri=path)
        if not hasattr(request, 'website'):
            request.website = self
        self.do_inbound(request)
        resource = resources.get(request)
        if return_request_too:
            return resource, request
        else:
            return resource


    def exec_simplate(self, path="/", request=None, response=None):
        """Given the URL path of a simplate, exec page two and return response.
        """
        resource, request = self.load_simplate(path, request, True)
        if response is None:
            response = Response(charset=self.charset_dynamic)
        context = resource.populate_context(request, response)
        exec resource.pages[1] in context  # let's let exceptions raise
        return response, context



import re


def by_regex(hook, regex_tuples, default=True):
    """A filter for hooks.

    regex_tuples is a list of (regex, filter?) where if the regex matches the
    requested URI, then the hook is applied or not based on if filter? is True
    or False.

    """
    regex_res = [ (re.compile(regex), disposition) \
                           for regex, disposition in regex_tuples.iteritems() ]
    def filtered_hook(request):
        for regex, disposition in regex_res:
            if regex.matches(request.line.uri):
                if disposition:
                    return hook(request)
                else:
                    return request
        if default:
            return hook(request)
        else:
            return request
    return filtered_hook


def by_dict(hook, truthdict, default=True):
    """A filter for hooks.

    truthdict is a mapping of URI -> filter? where if the requested URI is a
    key in the dict, then the hook is applied based on the filter? value.

    """
    def filtered_hook(request):
        do_hook = truthdict.get(request.line.uri, default)
        if do_hook:
            return hook(request)
        else:
            return request
    return filtered_hook



"""Implement re-execution of the aspen process.

When files change on the filesystem or we receive HUP, we want to re-execute
ourselves.

For thoughts on a more sophisticated approach, see:

    http://sync.in/aspen-reloading

"""
import os
import sys

import aspen


extras = set()
mtimes = {}


###############################################################################
# Thanks, Bob. :) #############################################################
# https://bitbucket.org/cherrypy/magicbus/src/41f5dfb95479/magicbus/wspbus.py #


# Here he saves the value of os.getcwd(), which, if he is imported early
# enough, will be the directory from which the startup script was run. This is
# needed by _do_execv(), to change back to the original directory before
# execv()ing a new process. This is a defense against the application having
# changed the current working directory (which could make sys.executable "not
# found" if sys.executable is a relative-path, and/or cause other problems).
_startup_cwd = os.getcwd()


try:
    import fcntl
except ImportError:
    max_cloexec_files = 0
else:
    try:
        max_cloexec_files = os.sysconf('SC_OPEN_MAX')
    except AttributeError:
        max_cloexec_files = 1024


def _do_execv():
    """Re-execute the current process.

    This must be called from the main thread, because certain platforms
    (OS X) don't allow execv to be called in a child thread very well.

    """
    args = sys.argv[:]
    aspen.log_dammit("Re-executing %s." % ' '.join(args))

    if sys.platform[:4] == 'java':
        from _systemrestart import SystemRestart
        raise SystemRestart
    else:
        args.insert(0, sys.executable)
        if sys.platform == 'win32':
            args = ['"%s"' % arg for arg in args]

        os.chdir(_startup_cwd)
        if max_cloexec_files:
            _set_cloexec()
        os.execv(sys.executable, args)


def _set_cloexec():
    """Set the CLOEXEC flag on all open files (except stdin/out/err).

    If self.max_cloexec_files is an integer (the default), then on
    platforms which support it, it represents the max open files setting
    for the operating system. This function will be called just before
    the process is restarted via os.execv() to prevent open files
    from persisting into the new process.

    Set self.max_cloexec_files to 0 to disable this behavior.

    """
    for fd in range(3, max_cloexec_files):  # skip stdin/out/err
        try:
            flags = fcntl.fcntl(fd, fcntl.F_GETFD)
        except IOError:
            continue
        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)

#
###############################################################################


execute = _do_execv

def clear_changes():
    global extras
    extras = set()

def if_changes(filename):
    extras.add(filename)


def check_one(filename):
    """Given a filename, return None or restart.
    """

    # The file may have been removed from the filesystem.
    # ===================================================

    if not os.path.isfile(filename):
        if filename in mtimes:
            aspen.log("File deleted: %s" % filename)
            execute()
        else:
            # We haven't seen the file before. It has probably been loaded
            # from a zip (egg) archive.
            return


    # Or not, in which case, check the modification time.
    # ===================================================

    mtime = os.stat(filename).st_mtime
    if filename not in mtimes: # first time we've seen it
        mtimes[filename] = mtime
    if mtime > mtimes[filename]:
        aspen.log("File changed: %s" % filename)
        execute()


def check_all():
    """See if any of our available modules have changed on the filesystem.
    """
    for name, module in sorted(sys.modules.items()):    # module files
        filepath = getattr(module, '__file__', None)
        if filepath is None:
            # We land here when a module is an attribute of another module
            # i.e., it exists twice in the sys.modules table, once as its
            # canonical representation, and again having been imported
            # within another module.
            continue
        filepath = filepath.endswith(".pyc") and filepath[:-1] or filepath
        check_one(filepath)

    for filepath in extras:                             # additional files
        check_one(filepath)


# Setup
# =====

def install(website):
    """Given a Website instance, start a loop over check_all.
    """
    for script_path in website.configuration_scripts:
        if_changes(script_path)
    website.network_engine.start_checking(check_all)



from aspen import Response
from aspen.resources.pagination import split_and_escape, Page
from aspen.resources.resource import Resource


class StringDefaultingList(list):
    def __getitem__(self, key):
        try:
            return list.__getitem__(self, key)
        except KeyError:
            return str(key)

ORDINALS = StringDefaultingList([ 'zero' , 'one' , 'two', 'three', 'four'
                                , 'five', 'six', 'seven', 'eight', 'nine'
                                 ])


class DynamicResource(Resource):
    """This is the base for JSON, negotiating, socket, and rendered resources.
    """

    min_pages = None  # set on subclass
    max_pages = None

    def __init__(self, *a, **kw):
        Resource.__init__(self, *a, **kw)
        pages = self.parse_into_pages(self.raw)
        self.pages = self.compile_pages(pages)


    def respond(self, request, response=None):
        """Given a Request and maybe a Response, return or raise a Response.
        """
        response = response or Response(charset=self.website.charset_dynamic)


        # Populate context.
        # =================

        context = self.populate_context(request, response)


        # Exec page two.
        # ==============

        try:
            exec self.pages[1] in context
        except Response, response:
            self.process_raised_response(response)
            raise


        # Hook.
        # =====

        try:
            response = self.get_response(context)
        except Response, response:
            self.process_raised_response(response)
            raise
        else:
            return response


    def populate_context(self, request, response):
        """Factored out to support testing.
        """
        context = request.context
        context.update(self.pages[0])
        context['request'] = request
        context['response'] = response
        context['resource'] = self
        return context


    def parse_into_pages(self, raw):
        """Given a bytestring, return a list of pages.

        Subclasses extend this to implement additional semantics.

        """

        pages = list(split_and_escape(raw))
        npages = len(pages)


        # Check for too few pages. This is a sanity check as get_resource_class
        # should guarantee this. Bug if it fails.
        assert npages >= self.min_pages, npages

        # Check for too many pages. This is user error.
        if self.max_pages is not None and npages > self.max_pages:
            type_name = self.__class__.__name__[:-len('resource')]
            msg = "%s resources must have at most %s pages; %s has %s."
            msg %= ( type_name
                   , ORDINALS[self.max_pages]
                   , self.fs
                   , ORDINALS[npages]
                    )
            raise SyntaxError(msg)

        return pages

    def compile_pages(self, pages):
        """Given a list of pages, replace the pages with objects.

        All dynamic resources compile the first two pages the same way. It's
        the third and following pages that differ, so we require subclasses to
        supply a method for that: compile_page.

        """

        # Exec the first page and compile the second.
        # ===========================================

        one, two = pages[:2]

        context = dict()
        context['__file__'] = self.fs
        context['website'] = self.website

        one = compile(one.padded_content, self.fs, 'exec')
        exec one in context    # mutate context
        one = context          # store it

        two = compile(two.padded_content, self.fs, 'exec')

        pages[:2] = (one, two)

        # Subclasses are responsible for the rest.
        # ========================================

        pages[2:] = (self.compile_page(page) for page in pages[2:])

        return pages

    @staticmethod
    def _prepend_empty_pages(pages, min_length):
        """Given a list of pages, and a min length, prepend blank pages to the
        list until it is at least as long as min_length
        """
        num_extra_pages = min_length - len(pages)
        #Note that range(x) returns an empty list if x < 1
        pages[0:0] = (Page('') for _ in range(num_extra_pages))

    # Hooks
    # =====

    def compile_page(self, *a):
        """Given a page, return an object.
        """
        raise NotImplementedError

    def process_raised_response(self, response):
        """Given a response object, mutate it as needed.
        """
        pass

    def get_response(self, context):
        """Given a context dictionary, return a Response object.
        """
        raise NotImplementedError
